unit ComboBoxPatch;
{$mode objfpc}{$H+}
interface

uses Windows, Messages, SysUtils, Classes, Controls, Graphics, LCLType, LCLIntf,
  Forms, LazUTF8, StdCtrls;

type
  EComboBoxPatchException = class(Exception);

  { TComboBoxPatch }

  TComboBoxPatch = class(TComponent)
  private
    FHintWnd: THintWindow;
    FCanvas: TCanvas;
    FAppEvnts: TApplicationProperties; // TApplicationEvents;
    FHintIndex: Integer;
    FActive: Boolean;
    FAlwaysShowHint: Boolean;
    FItemWidth: Integer;
    procedure AdjustHint(Wnd: HWND; Point: TPoint);
    procedure DoIdle(Sender: TObject; var Done: Boolean);
    procedure SetItemWidth(AValue: Integer);
    procedure ShowHint(Origin: TPoint; Text: String);
    procedure HideHint;
    procedure SetActive(const Value: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Active: Boolean read FActive write SetActive default true;
    property AlwaysShowHint: Boolean read FAlwaysShowHint write FAlwaysShowHint default false;
    property ItemWidth: Integer read FItemWidth write SetItemWidth;
  end;

const
  RSOnlyOneComboBoxPatch = 'Only one TComboBoxPatch can exist per application.';

var
  GlobalPatch: TComboBoxPatch = nil;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('CustomEdit',[TComboBoxPatch]);
end;

{ TComboBoxPatch }

procedure TComboBoxPatch.AdjustHint(Wnd: HWND; Point: TPoint);
var
  HintPt: TPoint;
  ItemRect: TRect;
  Item, TextWidth: Integer;
  buf: array of WCHAR;
  n, i: Integer;
  Str: String;
begin
  Windows.ScreenToClient(Wnd, Point);
  Item := SendMessage(Wnd, LB_ITEMFROMPOINT, 0, MakeLParam(Point.x, Point.y));
  if (Item and $00010000 <> $00010000) and (Item <> FHintIndex) then begin
    FHintIndex := Item;
    SendMessageW(Wnd, LB_GETTEXT, WParam(Item), LParam(LongInt(@buf)));
    n := SendMessageW(Wnd, LB_GETTEXTLEN, WParam(Item), LParam(0));
    SetLength(buf, n);
    SendMessageW(Wnd, LB_GETITEMRECT, Item, Integer(@ItemRect));
    FCanvas.Handle := GetDC(Wnd);
    FCanvas.Font := TComboBox(Self.Owner).Canvas.Font ;
    Str := '';
    for i := 0 to n - 1 do
      Str := Str + UnicodeChar(buf[i]);
    TextWidth := FCanvas.GetTextWidth(Str);
    ReleaseDC(Wnd, FCanvas.Handle);

    FCanvas.Handle := 0;
    if FAlwaysShowHint or ((not FAlwaysShowHint) and (ItemRect.Right < TextWidth)) then begin
      HintPt := ItemRect.TopLeft;
      Windows.ClientToScreen(Wnd, HintPt);
      ShowHint(HintPt, SysToUTF8(Str));
    end
    else begin
      HideHint;
    end;
    SetLength(buf, 0);
  end
  else begin
    if Item < 0 then HideHint;
  end;
end;

constructor TComboBoxPatch.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if Assigned(GlobalPatch) then begin
    raise EComboBoxPatchException.Create(RSOnlyOneComboBoxPatch);
  end
  else begin
    GlobalPatch:=Self;
  end;
  FHintIndex:=-1;
  FHintWnd:=THintWindow.Create(Self);
  FHintWnd.Color:=Application.HintColor;
  FCanvas:=TCanvas.Create;
  FAppEvnts := TApplicationProperties.Create(Self);
  FAlwaysShowHint:=false;
  Active:=true;
end;

destructor TComboBoxPatch.Destroy;
begin
  GlobalPatch:=nil;
  FCanvas.Free;
  FAppEvnts.Free;
  FHintWnd.Free;
  inherited Destroy;
end;

procedure TComboBoxPatch.DoIdle(Sender: TObject; var Done: Boolean);
var
  Pt: TPoint;
  buf: array[0..255] of CHAR;
  Wnd: HWND;
begin
  GetCursorPos(Pt);
  Wnd := WindowFromPoint(Pt);
  if Wnd <> 0 then begin
    GetClassName(Wnd, @buf, SizeOf(buf));
    if buf = 'ComboLBox' then begin
      AdjustHint(Wnd, Pt);
    end
    else begin
      HideHint;
    end;
  end;
end;

procedure TComboBoxPatch.SetItemWidth(AValue: Integer);
begin
  if FItemWidth = AValue then Exit;
  FItemWidth := AValue;
end;

procedure TComboBoxPatch.HideHint;
begin
  FHintWnd.ReleaseHandle;
  FHintIndex := -1;
end;

procedure TComboBoxPatch.SetActive(const Value: Boolean);
begin
  if Value <> FActive then begin
    FActive:=Value;
    if Value then
      FAppEvnts.OnIdle := @DoIdle
    else
      FAppEvnts.OnIdle := nil;
  end;
end;

procedure TComboBoxPatch.ShowHint(Origin: TPoint; Text: String);
var
  R: TRect;
begin
  Application.CancelHint;
  FHintWnd.Font.Size := 12;
  R := FHintWnd.CalcHintRect(1000,Text,nil);
  OffsetRect(R, Origin.X, Origin.Y);
  FHintWnd.ActivateHint(R, Text);
  SetWindowPos(FHintWnd.Handle, HWND_TOPMOST, R.Left, R.Top, FHintWnd.Width,
    FHintWnd.Height, SWP_NOACTIVATE);
end;

end.
